#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <winscard.h>

#define MAX_CONNECT (8)
#define TCP_PORT (6900)
#define FAIL_WAIT (500000)
#define INTERVAL_WAIT (100000)

SCARDCONTEXT sc = (SCARDCONTEXT)NULL;
SCARDHANDLE scd;

int loop_flag = 1;

void message_out(char *str)
{
#ifdef DEBUG
	printf(str);
#endif
}

void stop(int sig)
{
	message_out("Received signal.\n");
	loop_flag = 0;
}

int tcp_send(int s, void *buf, size_t len)
{
	fd_set write_set;
	struct timeval tv;
	int n, size, sent;

	if (s < 0) return -1;
	size = len;
	while (size) {
		tv.tv_sec = 0;
		tv.tv_usec = 0;

		FD_ZERO(&write_set);
		FD_SET(s, &write_set);
		n = select(s + 1, NULL, &write_set, NULL, &tv);
		if (n < 1) {
			if (n == 0) errno = ETIMEDOUT;
			return -1;
		}
		sent = send(s, buf, size, 0);
		if (sent == -1) return -1;

		buf = (char *)buf + sent;
		size -= sent;
	}

	return len;
}

int tcp_recv(int s, void *buf, size_t len)
{
	fd_set read_set;
	struct timeval tv;
	int n;

	tv.tv_sec = 0;
	tv.tv_usec = 0;

	FD_ZERO(&read_set);
	FD_SET(s, &read_set);
	n = select(s + 1, &read_set, NULL, NULL, &tv);
	if (n < 1) {
		if (n == 0) errno = ETIMEDOUT;
		return -1;
	}

	return recv(s, buf, len, 0);
}

int tcp_accept(int s, struct sockaddr *addr, int *addrlen)
{
	fd_set accept_set;
	struct timeval tv;
	int n;

	tv.tv_sec = 0;
	tv.tv_usec = 0;

	FD_ZERO(&accept_set);
	FD_SET(s, &accept_set);
	n = select(s + 1, &accept_set, NULL, NULL, &tv);
	if (n < 1) {
		if (n == 0) errno = ETIMEDOUT;
		return -1;
	}

	return accept(s, addr, addrlen);
}

void scard_open(void)
{
	LPTSTR rdr = NULL;
	DWORD cch = SCARD_AUTOALLOCATE, tmp;
	char *p;

	if (sc == (SCARDCONTEXT)NULL) return;
	if (SCardListReaders(sc, NULL, (LPTSTR)&rdr, &cch) != SCARD_S_SUCCESS) return;
	for (p = rdr; *p != '\0'; ) {
		if (SCardConnect(sc, p, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1, &scd, &tmp) == SCARD_S_SUCCESS) {
			BYTE recv[128];
			SCARD_IO_REQUEST sio = *SCARD_PCI_T1;
			DWORD rsz = sizeof(recv);
			if (SCardTransmit(scd, SCARD_PCI_T1, (BYTE *)"\x90\x30\x00\x00\x00", 5, &sio, recv, &rsz) == SCARD_S_SUCCESS) {
				if (rsz > 56 && recv[4] == '\x21' && recv[5] == '\0') {
					message_out("B-CAS card found.\n");
					break;
				}
			}
		}
		SCardDisconnect(scd, SCARD_LEAVE_CARD);
		scd = (SCARDHANDLE)NULL;
		p += strlen(p) + 1;
	}
	SCardFreeMemory(sc, rdr);
}

void scard_init(void)
{
	if (SCardEstablishContext(SCARD_SCOPE_USER, NULL, NULL, &sc) != SCARD_S_SUCCESS) {
		sc = (SCARDCONTEXT)NULL;
		return;
	}
	scard_open();
}

void scard_finish(void)
{
	if (scd != (SCARDCONTEXT)NULL) SCardDisconnect(scd, SCARD_LEAVE_CARD);
	SCardReleaseContext(sc);
	sc = (SCARDCONTEXT)NULL;
}

int main(int argc, char *argv[])
{
	int tr_sock[MAX_CONNECT];
	struct sockaddr_in addr;
	int sock1, sock2, adlen, n, l;

	signal(SIGINT, stop);
	signal(SIGHUP, stop);
	signal(SIGTERM, stop);
	signal(SIGPIPE, SIG_IGN);

	if ((sock1 = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr, "Can't attach socket interface.\n");
		return EXIT_FAILURE;
	}
	message_out("Socket created...\n");

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(TCP_PORT);
	addr.sin_addr.s_addr = INADDR_ANY;
	if (bind(sock1, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr, "Can't bind.\n");
		close(sock1);
		return EXIT_FAILURE;
	}
	if (listen(sock1, 1)) {
		fprintf(stderr, "Can't listen.\n");
		close(sock1);
		return EXIT_FAILURE;
	}
	message_out("Listening OK...\n");

#ifndef DEBUG
	daemon();
#endif
	scard_init();
	if (sc == (SCARDCONTEXT)NULL) {
		fprintf(stderr, "Can't attach smartcard interface.\n");
		close(sock1);
		return EXIT_FAILURE;
	}
	for (n = 0; n < MAX_CONNECT; n++) tr_sock[n] = 0;

	while (loop_flag) {
		if (scd == (SCARDHANDLE)NULL) {
			scard_open();
			if (scd == (SCARDHANDLE)NULL) usleep(FAIL_WAIT);
		}
		for (n = 0; n < MAX_CONNECT; n++) {
			char send[256];
			int s = tr_sock[n];
			if (s == 0) continue;
			if ((l = tcp_recv(s, send, 256)) < 0) continue;
			if (l == 0) {
				close(s);
				tr_sock[n] = 0;
				message_out("Connection close\n");
				continue;
			}
			if (scd == (SCARDHANDLE)NULL) {
				message_out("error send\n");
				tcp_send(s, "\0", 1);
				continue;
			} else {
				BYTE recv[128];
				SCARD_IO_REQUEST sio = *SCARD_PCI_T1;
				DWORD rsz = sizeof(recv) - 1;
				WORD len = *(BYTE *)send;
				LONG rst;
				if ((rst = SCardTransmit(scd, SCARD_PCI_T1, send+1, len, &sio, recv+1, &rsz)) != SCARD_S_SUCCESS) {
					SCardDisconnect(scd, SCARD_LEAVE_CARD);
					scd = (SCARDHANDLE)NULL;
					rsz = 0;
				}
				recv[0] = rsz;
				if (tcp_send(s, recv, rsz+1) < 0) {
					close(s);
					tr_sock[n] = 0;
				}
			}
		}
		adlen = sizeof(addr);
		if ((sock2 = tcp_accept(sock1, (struct sockaddr *)&addr, &adlen)) > 0) {
			for (n = 0; n < MAX_CONNECT; n++) {
				if (tr_sock[n] != 0) continue;
				message_out("Connect...\n");
				tr_sock[n] = sock2;
				break;
			}
			if (n >= MAX_CONNECT) {
				close(sock2);
				message_out("Force connection close(restriction)...\n");
			}
		}
		usleep(INTERVAL_WAIT);
	}
	scard_finish();

	for (n = 0; n < MAX_CONNECT; n++) {
		int s = tr_sock[n];
		if (s == 0) continue;
		shutdown(s, SHUT_RDWR);
		close(s);
	}
	shutdown(sock1, SHUT_RDWR);
	close(sock1);
	message_out("Socket closed...\n");

	return EXIT_SUCCESS;
}
